#
# Licensed Materials - Property of IBM
# Rational ClearQuest
# (C) Copyright IBM Corp. 2006, 2014. All Rights Reserved.
#
# US Government Users Restricted Rights -
# Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#

use Getopt::Long;
use Time::Local;
use File::Temp "tmpnam";
use FindBin;
use lib "$FindBin::Bin";
use strict;

#
# This script updates the history table, converting db server local times to
# UTC and updating the comments field to indicate what changes were made.
# It was written as part of the solution to RATLC00718629.
#
# It figures out what changes are needed by combining information read
# from the history, ratl_id_blocks, and ratl_replicas tables with a
# configuration file specified by the user.  It potentially updates 3
# columns in the history table:
#   action_timestamp, expired_timestamp and comments.
#
# It can generate either a pdsql script consisting of update
# statements, or a file containing tab separated values that is then
# passed to a hidden pdsql command written specifically for use by
# this script.  This hidden command does block updates of the history
# table.  The former is easier to understand but the latter runs about
# 30x faster, so guess which one is the default.
#
# This script has 3 stages:
# 1) Read the database
# 2) Analyze what was read, and
# 2a) Produce repair commands, if needed, or
# 2b) Just report on whether any repair is needed
# 3) Run the repair commands
#
# It uses pdsql to read and update the database.  There are assumptions
# made about the format of the output of reading the database that might
# not be valid if, for example, sqlplus were used instead.
#
# It uses this query to read the history table:
#
# select dbid, entity_dbid, action_timestamp, expired_timestamp,
#   entitydef_name, comments from history
#   order by entitydef_id, entity_dbid, action_timestamp desc, dbid desc
#
# You have to include either entitydef_id or entitydef_name in the
# sort because stateless and stateful records can have the same
# entity_dbids.  Sorting by entitydef_id instead of entitydef_name is
# used on the theory that sorting on a numeric column will be faster
# than sorting on a varchar column.  entitydef_name is selected so it
# can be used diagnostic messages, but is otherwise unused.  The last
# sort criterion is dbid desc to break ties when action_timestamps are
# identical.
#
# It reads ratl_id_blocks and ratl_replica with this query:
#
# select t1.base, t1.limit, t2.name, t1.type, t1.alloc_tm from
#   ratl_id_blocks t1, ratl_replicas t2 where
#   t1.type = 3 and t1.replica = t2.dbid order by t2.name, t1.base
#
# Type is always 3 (aux series), but can be used to check that the
# query was run correctly.  alloc_tm is selected so that the ranges
# specified by the config file can be sanity checked, although it's
# not clear how much good this will do in practice. Some sample dbs
# may have weird times in them already, but gross errors could be
# caught.  For example, if the alloc_tm for the only block for a site
# is 2003, while the config file includes dates from 2001 through 2002
# for that site, this is almost certainly an error.  But you might want
# to use the same conversion for all dates at a site and specify a very
# low start date, or high end date, snd we don't want to be overly
# picky.

# Notes:
# 1) Sitenames are upper cased.
# 2) In order to prevent applying already generated sql commands to the
#    wrong user database, which would probably require a database restore
#    to recover from, it does sanity checking in stages 2 and 3.  In stage 2
#    it makes sure that the idblock and history files were read from the
#    same database.  In stage 3, it checks to see that the db being repaired
#    is in the same family as the one the repair commands were generated for.
# 3) Files are used to communicate between the stages of the processing.
#    Although bi-directional pipes would make more sense in some cases,
#    they are problematic on windows.  Plus, the history file cannot be
#    read into memory since it will be too big if the history table is
#    large and/or memory is small.  Plus, since each stage can be run
#    individually, files need to be used anyway.

# command line options
my $g_help = 0;          # Print usage message and exit
my $g_cfgfile = "";      # Config file with info on how to adjust times
                         # by site and date 
my $g_fmt = "";          # Output format - legal values are "s" and "v",
                         # where s == sql update statements and v == values
my $g_verb = 1;          # Verbose mode - 0 means off, non-0 means print
                         # warnings and print info if level is <= $g_verb
my $g_sqlfile = "";      # Output of this script - contains either sql to 
                         # fix history or tab separated values
my $g_idblockfile = "";  # Output of select from ratl_id_blocks table, input
                         # to analysis phase
my $g_histfile = "";     # Output of select from history table, input
                         # to analysis phase
my $g_pdsqlargs = "";    # Arguments to pdsql, passed as 1 command line argument
                         # in quotes
my $g_cmdarg = "";       # readdb, checkdb, gensql, runsql, fixdb
my $g_removeutc = 0;     # reverse utc adjustments made in the past
# end of command line options

my $g_dbvendor = "";        # Derived from $g_pdsqlargs, used to generate vendor specific sql
my $g_idblock_dbinfo = "";  # Database info selected along with id block info
my $g_history_dbinfo = "";  # Database info selected along with the history table
my $g_sql_dbinfo = "";      # Database info passed along in the generated sql
my $g_fixup_dbinfo = "";    # Database info read from the db being updated
my $MATCHES_NEEDED = 4;   # The number of fields that need to match
                          # between $g_sql_dbinfo and $g_fixup_dbinfo in order
                          # to allow the update to proceed.

my $FMT_VALUES = "v"; # generate tab separated list of values
my $FMT_SQL = "s";    # generate sql update statements

my $DEF_INFO_VERB = 2; # default verbosity for info messages

# vars that say what to do.  docheckdb and dogensql are mutually exclusive,
# as are docheckdb and dorunsql.  These are set based on the value of $g_cmdarg.
my $g_doreaddb = 0;
my $g_docheckdb = 0;
my $g_dogensql = 0;
my $g_dorunsql = 0;
my $g_doverifycfg = 0;
my $g_dotest = 0;        # run some internal "unit" tests if non-zero.

my $UNINIT = -1;       # uninitialized numeric value
my $NULLSTR = "NULL";  # pdsql returns NULL for nulls
my $BOGUS = "Bogus";   # error indicator
my $BOGUSNUM = 2**33;  # perl uses floats for numbers so we can make
                       # this > 2**32.  This makes it bigger than any
                       # dbid or time_t.
my $IDTODBID = 2**25;
my $AUXTYPE = 3;
my $MAX_ADJ = 14 * 3600; # maximum a time can be adjusted, in seconds

my %g_sitehash;          # maps site names and dates to time adjustments
my @g_idblocks = ();     # maps ids to sites

# start of vars used as indexes into the rows of the history table
my $DBID_IDX = 0;
my $ENTITY_DBID_IDX = 1;
my $ACTION_TIMESTAMP_IDX = 2;
my $EXPIRED_TIMESTAMP_IDX = 3;
my $ENTITYDEF_NAME_IDX = 4;
my $COMMENTS_IDX = 5;
# end of vars used as history indexes
# start of vars used as indexes to record extra info about each row in
# the history file
my $LINENO_IDX = 6;       # line number in file, used in error and warning messages
my $ACT_TS_COMM_IDX = 7;  # conventional comment for action_timestamp
my $ROW_CHANGES_IDX = 8;  # any changes to this row?
my $COMMENTS_FMT_IDX = 9; # what is in the comments field for a row?
# end of vars used to record extra info

# start of vars used as indexes into the rows of the site descriptions in %g_sitehash
my $STARTTIME_IDX = 0;
my $ENDTIME_IDX = 1;
my $ADJINSECS_IDX = 2;
my $ADJOFFSET_IDX = 3;
# next 2 just used for diagnostic messages when processing config file
my $STARTTIME_TS_IDX = 4;
my $ENDTIME_TS_IDX = 5;
#
# start of vars used as indexes into the rows of the idblocks array
my $BASE_IDX = 0;
my $LIMIT_IDX = 1;
my $SITENAME_IDX = 2;
my $TYPE_IDX = 3;
my $ALLOC_TM_IDX = 4;
#

# constant vars used in the bit mask that indicates what changes were
# made to a row
my $ACT_TS_CHANGE = 1;        # action timestamp changed
my $EXP_TS_CHANGE = 2;        # expired timestamp changed
my $COMM_CHANGE = 4;          # comments changed
# end of bit mask constants

# Bitmap constants that describe the comments field
my $NO_COMM = 0;
my $ACT_COMM = 1;
my $EXP_COMM = 2;
my $BOGUS_COMM = 4;

# end of constants that describe the comments field

# commit after every 100 updates - only meaningful if generating update
# statements
my $COMMIT_FREQ = 100;
# count of update statements since start of transaction
my $g_update_count = 0;

# number of updates needed for the history table
my $g_rowschanged = 0;
# Next 4 hashes used to print info about what changed.  Keys are record names,
# values are as described.
my %g_updated_recs;      # number of records that needed changes
my %g_updated_rows;      # number of rows in history table being updated
my %g_nochanges_recs;    # number of records that did not need changes
my %g_nochanges_rows;    # number of rows in history table not being updated
# Key is sitename, value is number of history rows with action_timestamps
# that need fixing.
my %g_sitefixups;

main();
exit(0);

sub main {
    if (!GetOptions("verbose=i" => \$g_verb,
                    "fmt=s" => \$g_fmt,
                    "help" => \$g_help,
                    "sqlfile=s" => \$g_sqlfile,
                    "idblockfile=s" => \$g_idblockfile,
                    "historyfile=s" => \$g_histfile,
                    "pdsqlargs=s" => \$g_pdsqlargs,
                    "cmd=s" => \$g_cmdarg,
                    "removeutc" => \$g_removeutc,
                    "cfg=s" => \$g_cfgfile)) {
        usage();
    }

    # checkArgs processes the config file, so $g_doverifycfg has been
    # handled already
    checkArgs();
    my $stat = 0;
    if ($g_doreaddb) {
        $stat = readFromDB();
        checkStat($stat, "error in readFromDB");
    }
    if ($stat == 0 && ($g_docheckdb || $g_dogensql)) {
        # first read ratl_id_blocks query output
        $stat = readIdBlockFile($g_idblockfile);
        checkStat($stat, "error in readIdBlockFile");
        $stat = checkIdBlockConfig();
        checkStat($stat, "id blocks and config file are inconsistent");
        
        # we already know that at most 1 of these is true
        $stat = checkAndGen($g_dogensql, $g_histfile, $g_sqlfile, undef);
        checkStat($stat, "error in checkAndGen");
    }
    if ($stat == 0 && $g_dorunsql) {
        if ($g_dogensql && $g_rowschanged == 0) {
            prtmsg("Repair step skipped because no repair sql was generated");
        } else {
            $stat = repairDB();
            checkStat($stat, "error in repairDB");
        }
    }
    reportResults();
    exit($stat);
}

sub checkStat {
    my $stat = shift;
    my $msg = shift;
    if ($stat) {
        err("$msg:  return code of $stat");
    }
}

# Note test command omitted from usage message.  Should it be?
sub usage {
    my $err = shift;
    if ($err) {
        print STDERR "Error:  $err\n";
    }
    my $msg = "Usage: cqperl $0\n";
    $msg .= <<'EOS';
                 -cmd readdb|checkdb|gensql|runsql|fixdb|verifycfg
                 [-help]
                 [-verbose n]
                 [-fmt s|v]
                 [-historyfile filename]
                 [-sqlfile filename]
                 [-idblockfile filename]
                 [-pdsqlargs "args"]
                 [-cfg configfile]
                 [-removeutc]

  -help prints this message
  -verbose sets the verbosity level - higher means users see more messages
  -cfg specifies the configuration file name that contains the information for
       mapping sites to time zones
  -historyfile specifies the name of the file where history records are 
       written to and read from
  -idblockfile specifies the name of the file where ratl_id_blocks records are 
       written to and read from
  -sqlfile specifies the name of the file where repair commands are
       written to and read from
  -pdsqlargs specifies the command line arguments to be used with pdsql - they
       must be quoted appropriately for the command line processor
       being used (cmd, tcsh, bash, etc.)
  -removeutc says to convert timestamps to database server local time.
       This should only be used in an emergency.
  -cmd says what the program should do.  Legal values are:
       readdb    - read the history and ratl_id_blocks tables from
                   the database, write results to files
       checkdb   - read the files produced by readdb, report whether
                   any repair is needed
       gensql    - same as checkdb but also generate repair commands
                   if needed
       runsql    - run previously generated repair commands
       fixdb     - readdb, gensql and runsql
       verifycfg - check the validity of the config file and exit
       These can be specified as a comma separated list.  For example
       -cmd readdb,checkdb

  File arguments can be input, output or both depending on the -cmd option.
  For example, readdb will only write files, runsql will only read them,
  and fixdb will both write and read.
  Some arguments are required for some commands, but optional for others.  For
  example, you do not need -cfg for readdb or runsql, but it is required
  for checkdb, gensql, fixdb and verifycfg.

Examples:
  Note that these commands need to be entered on one line.

  Read the history and id block information from the database.
  cqperl  timechange.pl -cmd readdb -hist c:\temp\history.out 
    -idbl c:\temp\idblocks.out 
    -pdsqlargs "-v access -u admin -db c:\cqtest\baltic_s1_sampl.mdb"
  
  Generate repair commands, if needed.
  cqperl  timechange.pl -cmd gensql -hist c:\temp\history.out 
    -idbl c:\temp\idblocks.out -sqlfile c:\temp\fixdb.sql -cfg c:\temp\cfg.in
  
  Execute the repair commands.
  cqperl  timechange.pl -cmd runsql -sqlfile c:\temp\fixdb.sql 
    -pdsqlargs "-v access -u admin -db c:\cqtest\baltic_s1_sampl.mdb"

  Do the 3 previous steps in one command.
  cqperl  timechange.pl -cmd fixdb -hist c:\temp\history.out 
    -idbl c:\temp\idblocks.out -sqlfile c:\temp\fixdb.sql 
    -cfg c:\temp\cfg.in 
    -pdsqlargs "-v access -u admin -db c:\cqtest\baltic_s1_sampl.mdb"
  
  Just check to see whether any history records need updating.
  cqperl  timechange.pl -cmd readdb,checkdb -hist c:\temp\history.out 
    -idbl c:\temp\idblocks.out -cfg c:\temp\cfg.in
    -pdsqlargs "-v access -u admin -db c:\cqtest\baltic_s1_sampl.mdb"

  The pdsqlargs will vary depending on the database vendor.  See
  the pdsql manual for more details.  
EOS
    print STDERR "$msg\n";
    # Since the usage message is so long, repeat error message at the end.
    if ($err) {
        print STDERR "Error:  $err\n";
    }
    exit(2);
}

# Sort by first elements of the arrays that represent the time ranges for
# each site.
sub sitesort {
    return $$a[0] <=> $$b[0];
}

sub prtmsg {
    my $msg = shift;
    print STDERR "$msg\n";
}

# usually fatal
sub err {
    my $msg = shift;
    my $nonfatal = shift;
    prtmsg("Error:  $msg");
    if (!$nonfatal) {
        exit(1);
    }
}

# Some problem needs to be fixed, but not fatal.
sub warning {
    if ($g_verb) {
        my $msg = shift;
        prtmsg("Warning:  $msg");
    }
}

# Mostly for debugging.  
sub info {
    if ($g_verb) {
        my $msg = shift;
        my $level = shift;
        if (! defined $level) {
            $level = $DEF_INFO_VERB;
        }
        if ($level <= $g_verb) {
            prtmsg("Info:  $msg");
        }
    }
}

# Note that when autocommit mode is off, transactions are implicitly
# started, according to the odbc manual.
sub set_trans_mode {
    if ($g_fmt eq $FMT_SQL) {
        my $outputhandle = shift;
        print $outputhandle "autocommit_off;\n";
    }
}

sub commit_trans {
    if ($g_fmt eq $FMT_SQL) {
        if ($g_update_count > 0) {
            my $outputhandle = shift;
            print $outputhandle "commit;\n";
            $g_update_count = 0;
        }
    }
}

# All dates coming into the tool are character strings.  They are
# converted to seconds, manipulated, and then converted back to
# character strings.  In principle, it does not matter what timezone
# is used when doing the conversions.  But there are discontinuities at
# daylight savings time switchovers for some locales.  Since we cannot
# resolve the inherent ambiguities in the hour between 1 am and 2 am
# in the fall, we might as well just use gmt as the conversion locale.
# Convert timestamp to seconds since epoch.
# Handles both 2005-10-31 02:00:00 and just 2005-10-31
sub ts_to_seconds {
    my $ts = shift;
    # convert multiple white space chars to single blanks to keep split happy
    $ts =~ s/\s+/ /g;
    my $seconds = $BOGUSNUM;
    if ($ts =~ m/^\d\d\d\d-\d\d-\d\d( \d\d:\d\d:\d\d)?$/) {
        my @tmarr = split(/[- :]/, $ts);
        # months are 0-11 in perl, years are since 1900
        $seconds = timegm($tmarr[5], $tmarr[4], $tmarr[3],
                             $tmarr[2], $tmarr[1] - 1, $tmarr[0] - 1900); 
    }
    return $seconds;
}

# Converts seconds to timestamp.
sub seconds_to_ts {
    my $secs = shift;
    if ($g_verb >= 3) {
        info(scalar(gmtime($secs)));
    }
    my @timearr = gmtime($secs);
    # + 1900 because years are since 1900, + 1 because months are 0 based.
    my $ts = sprintf("%4d-%02d-%02d %02d:%02d:%02d",
                            $timearr[5] + 1900, $timearr[4] + 1,
                            $timearr[3], $timearr[2],
                            $timearr[1], $timearr[0]);
    return $ts;
}

# Convert seconds to [+-]HH:MM
# Used during scanning of config file to normalize offset representation.
sub secs_to_hours_minutes {
    my $secs = shift;
    my $adj;
    # Timezone offsets can go from -12 to +14.  But since removeutc mode
    # needs to be able to removeutc whatever was done in the past, we need
    # up to 14 hours in both directions.
    if ($secs > $MAX_ADJ || $secs < -$MAX_ADJ) {
        $adj = $BOGUS;
    } else {
        my $neg = 0;
        if ($secs < 0) {
            $neg = 1;
            $secs = -$secs;
        }
        my $h = $secs / 3600;
        my $m = ($secs % 3600) / 60 ;
        $adj = sprintf("%s%02d:%02d", ($neg ? "-" : "+"), $h, $m);
    }
    return $adj;
}

# convert timezone adjustment to seconds
# 6:30 means 6 hours 30 minutes, as does +6.5
# Valid adjustments are -2.5, 02:30, -2:00, 2
sub timezone_adj_to_seconds {
    my $adj = shift;
    # remove leading +
    $adj =~ s/^\+//;
    my $amt = $BOGUSNUM;
    if ($adj =~ m/^[-]?\d{1,2}([\.:]\d{1,2})?$/) {
        my @adjarr = split(/:/, $adj);
        my $neg = ($adjarr[0] =~ s/-//);
        $amt = (3600 * $adjarr[0] + 60 * $adjarr[1]) * ($neg ? -1 : 1);
    }
    return $amt;
}

# Subtract adjustment from timestamp to get UTC from local time,
# return result.  Subtract because the timezone convention is to
# specify timezones west of Greenwich as negative, but you have to
# subtract that amount to get to utc. E.G.
# 2005-11-09 14:38:14 est - (-5:00) => 2005-11-09 19:38:14 utc
# cqperl does not ship with the cpan Date modules, which allegedly
# provide date manipulation functionality, so I'm using Time::Local.
sub fixtime_for_dbid {
    my $dbid = shift;
    my $act_ts = shift;
    my $sitename = "";

    # First find the site where this dbid was allocated.  This could
    # be pretty slow for a site with hundreds of id blocks and millions
    # of history records, but not as slow as before it was using binary
    # search.
    my $idx;
    my $cnt;
    ($idx, $cnt) = brangesrch(\@g_idblocks, $dbid);
    if ($idx == -1) {
        err("Failed to find $dbid in an id block, count = $cnt");
    }
    my $idbl = $g_idblocks[$idx];
    $sitename = $$idbl[$SITENAME_IDX];

    # This error should only occur if the query used to read ratl_id_blocks
    # was incorrect.
    if (!$sitename) {
        err("dbid to sitename lookup failed for $dbid");
    }

    # These errors indicate some problem in the configuration file.  For
    # example, if a site was omitted, or the wrong name used.
    my $adjarr = $g_sitehash{$sitename};
    if (! defined $adjarr) {
        err("no conversion info for $sitename, dbid = $dbid");
    }

    # Now find the time period when this dbid was allocated.
    # Binary search doesn't buy so much here since these arrays will
    # tend to be small, maybe just 1 per site, or more likely, 2 entries
    # per year, one for dst and one for standard time.
    my $secs = ts_to_seconds($act_ts);
    ($idx, $cnt) = brangesrch($adjarr, $secs);
    if ($idx == -1) {
        err("Did not find mapping of $act_ts at site $sitename, count = $cnt");
        return (0, $BOGUS);
    }
    my $interval = $$adjarr[$idx];
    $g_sitefixups{$sitename}++;
    my $adjsecs = $$interval[$ADJINSECS_IDX];
    my $adjcomm = $$interval[$ADJOFFSET_IDX];
    # see comment above for why this is a subtraction
    my $newsecs = $secs - $adjsecs;
    my $adjact_ts = seconds_to_ts($newsecs);
    return ($adjact_ts, $adjcomm);
}

sub remove_utc {
    my $comm = shift;
    my $act_ts = shift;
    my $actoffset = extractActionOffset($comm);
    my $adj_secs = timezone_adj_to_seconds($actoffset);
    my $secs = ts_to_seconds($act_ts);
    # undo utc adjustment by adding, rather than subtracting
    my $newsecs = $secs + $adj_secs;
    my $new_ts = seconds_to_ts($newsecs);
    return $new_ts;
    
}

# Sort an array of history records.  Called after action_timestamps have
# been fixed for a particular entity.
# This sort has to mirror the one used in the select statement.  In particular,
# it breaks ties on action_timestamp by using dbids.  This will not work in all
# cases, since you could have 2 actions occur within a second of each other, and
# have a block switch occur between them, and have the block switched to
# have lower ids.  This is presumably an infrequent occurrance, but it could
# happen occasionally in big dbs.
sub ts_sort {
    return $$b[$ACTION_TIMESTAMP_IDX] cmp $$a[$ACTION_TIMESTAMP_IDX] ||
        $$b[$DBID_IDX] <=> $$a[$DBID_IDX];
}

# See whether existing action and expired timestamps match as expected.
sub check_ts_consistency {
    my $act_ts = shift;
    my $exp_ts = shift;
    if ($act_ts ne $exp_ts) {
        # Should do a fuzzy match.
        return 0;
    }
    return 1;
}


# Make sure that command line input file args are OK
sub checkInputFile {
    my $fn = shift;
    my $msg = shift;
    if (!$fn) {
        usage("Input file name missing:  $msg");
    } elsif (!(-e $fn && -r $fn)) {
        usage("$fn does not exist or is not readable");
    }
}

# Make sure that command line output file args are OK
sub checkOutputFile {
    my $fn = shift;
    my $msg = shift;
    if (!$fn) {
        usage("Output file name missing:  $msg");
    } elsif (-e $fn && (! -w $fn)) {
        usage("$fn exists but is not writable");
    }
}

# We already know this file exists and is readable, so expect no errors on
# the open.  Create mapping of sites to time adjustments.
# Format of Site lines is:
# Site: name offset [start time] - [end time]
#
# Site:	Site2	+6:30
# Site:	Site3	-5 2005-10-31 02:00:00 - 2006-04-01 02:00:00
# Site:	Site3   -5 2004-10-25 - 2005-03-28
#
# Creates a hash of arrays with site names as keys.  Array elements are
# refs to arrays that contain 3 elements - start time, end time and amount
# to adjust.  Start and end times are in seconds cince the epoch, adjustment
# is in seconds to subtract from a time stamp to convert it to utc.
#
sub readConfigFile {
    my $cfile = shift;
    my $errs = shift;
    # null out sitehash so that this can be called multiple times in one
    # test run.
    %g_sitehash = ();
    my $errlim = 20;
    open CF, "<$cfile" or die "Cannot open $cfile for reading";
    while (<CF>) {
        # skip comments - put next inside a block because otherwise
        # emacs indentation gets all confused by the # in the expression
        if (m/^\s*#/) {
            next;
        }
        # skip blank lines
        next if (m/^\s*$/);
        chomp;
        # trim white space
        s/^\s*//;
        s/\s*$//;
        if (s/^site:\s*//i) {
            my @sitespec = split(" ", $_, 3);
            # The array must contain at least 2 elements - sitename
            # and adjustment.  The rest can be optional.  The minutes
            # part may be omitted.
            # 0 - sitename
            # 1 - amount to adjust by
            # 2 - rest of line
            my $sitename = uc($sitespec[0]);
            # we are not doing this site
            my $starttime = 0;
            my $endtime = 2**31 - 1;
            my $adjinsecs = 0;
            my $adjoffset = "";
            if (lc($sitespec[1]) eq "skip") {
                push @$errs, "skip directive not implemented";
                # If the file contains both skip and some other specs
                # for a site, this value will trigger errors in the
                # consistency check, although they may not be reported
                # very well.
                $adjinsecs = $BOGUSNUM;
                # should not make it into the generated sql!
                $adjoffset = $BOGUS;
            } else {
                # convert adjustment to seconds
                $adjinsecs = timezone_adj_to_seconds($sitespec[1]);
                if ($adjinsecs == $BOGUSNUM) {
                    push(@$errs, "Timezone offset error at line $. in config file\n  Could not convert $sitespec[1] to seconds");
                }
                # normalize for use in comments
                $adjoffset = secs_to_hours_minutes($adjinsecs);
                if ($adjoffset eq $BOGUS) {
                    push(@$errs, "Timezone offset error at line $. in config file\n  Could not convert $adjinsecs to a valid offset");
                }
                if ($sitespec[2]) {
                    my @tsarr = split(/\s+-\s+/, $sitespec[2]);
                    if ($#tsarr != 1) {
                        push(@$errs, "Syntax error in line $. in config file\n  $_");
                        next;
                    }
                    # normalize representation
                    if (! ($tsarr[0] =~ m/\s/)) {
                        $tsarr[0] .= " 00:00:00";
                    }
                    if (! ($tsarr[1] =~ m/\s/)) {
                        $tsarr[1] .= " 00:00:00";
                    }
                    $starttime = ts_to_seconds($tsarr[0]);
                    $endtime = ts_to_seconds($tsarr[1]);
                    if ($starttime == $BOGUSNUM) {
                        push(@$errs, "Error converting $tsarr[0] to seconds at line $.");
                        next;
                    } elsif ($endtime == $BOGUSNUM) {
                        push(@$errs, "Error converting $tsarr[1] to seconds at line $.");
                        next;
                    }
                    if ($endtime <= $starttime) {
                        push(@$errs, "End time <= start time at line $.\n  $_");
                    }
                }
            }
            my $ar = $g_sitehash{$sitename};
            # $adjinsecs used for adjustment, $adjoffset used in
            # conventional comments
            # If the order changes here, the values of the index variables
            # must change, too.  E.g. $STARTTIME_IDX
            push(@$ar, [$starttime, $endtime, $adjinsecs, $adjoffset,
                        seconds_to_ts($starttime), seconds_to_ts($endtime)]);
            $g_sitehash{$sitename} = $ar;
            # make sure each site has an entry in the array
            $g_sitefixups{$sitename} = 0;
        } else {
            push(@$errs, "Unrecognized config file directive at line $.\n  $_");
        }
        if ($#$errs >= $errlim) {
            unshift(@$errs, "Config file error limit of $errlim exceeded");
            last;
        }
    }
    close CF;

    if (scalar(keys %g_sitehash) > 0) {
        for my $k (sort keys %g_sitehash) {
            my $ar = $g_sitehash{$k};
            # sort the entries for each site by starttime
            @$ar = sort sitesort @$ar;
            info("$k", 2);
            for my $e (@$ar) {
                info("  " . join(" ", @$e), 2);
            }
            # check consistency
            for (my $i = 0; $i < $#$ar; $i++) {
                # Allow equality or 1 second difference
                if ($$ar[$i][$ENDTIME_IDX] < $$ar[$i+1][$STARTTIME_IDX] - 1) {
                    push @$errs, "Gap in times for site $k between $$ar[$i][$ENDTIME_TS_IDX] and $$ar[$i+1][$STARTTIME_TS_IDX]";
                } elsif ($$ar[$i][$ENDTIME_IDX] > $$ar[$i+1][$STARTTIME_IDX]) {
                    push @$errs, "Overlap in times for site $k:  $$ar[$i][$ENDTIME_TS_IDX] > $$ar[$i+1][$STARTTIME_TS_IDX]";
                }
            }
        }
    } else {
        push @$errs, "No site information processed";
    }

    if ($#$errs >= 0) {
        return 1;
    }
    return 0;
}

# We already know this file exists and is readable.  Create mapping of
# dbids to sites.
sub readIdBlockFile {
    my $idbfile = shift;
    open IDB, "<$idbfile" or die "Cannot open $idbfile for reading: $!\n";
    my $saw_hdrs = 0;
    while (<IDB>) {
        next if (m/^>/);     # pdsql prompt lines
        next if (m/^\s*$/);  # blank lines
        chomp;
        s/^\s+//;  # strip leading whitespace
        s/\s+$//;  # strip trailing whitespace
        s/\s+/ /g; # turn all white space into single blanks
        # skip column headers and store dbinfo
        if ($saw_hdrs == 0) {
            $saw_hdrs = 1;
            next;
        } elsif ($saw_hdrs == 1) {
            if ($g_idblock_dbinfo eq "") {
                ($g_idblock_dbinfo = $_) =~ s/\s+/ /g;
                if (!checkDbInfo($g_idblock_dbinfo)) {
                    err("Database info in wrong format in $idbfile:  $g_idblock_dbinfo");
                }
            } else {
                $saw_hdrs = 2;
            }
            next;
        }

        # Since alloc_tm has an embedded blank between the day and
        # time parts, it should be the last field in the select.  The
        # limit of 5 on the split means that the 5th field will get
        # whatever is left at the end of the line.  Should we worry about
        # it being null??
        my @idblock = split(" ", $_, 5);
        # make sure this is an aux series block
        if ($idblock[$TYPE_IDX] != $AUXTYPE) {
            err("Bogus idblock type $idblock[$TYPE_IDX] at line $. of $idbfile");
        }
        # site names are upper case
        my $sitename = uc($idblock[$SITENAME_IDX]);
        # Add 2 ** 25 to values in the dbid table.
        # Subtract 1 from the limit so that the blocks don't overlap - conforms
        # to brangesrch convention of including at both ends of a range.
        push (@g_idblocks, [$idblock[$BASE_IDX] + $IDTODBID,
                          $idblock[$LIMIT_IDX] + $IDTODBID - 1,
                          $sitename, $idblock[$TYPE_IDX],
                          $idblock[$ALLOC_TM_IDX]]);
    }
    close IDB;
    # Wouldn't have to do the sort if the select used the right order by
    # clause in the first place ....
    @g_idblocks = sort sitesort @g_idblocks;
    for my $e (@g_idblocks) {
        info(join(", ", @$e), 2);
    }
    # no error
    return 0;
}

# Check to see that the config file is not grossly inconsistent with the
# idblock file.
sub checkIdBlockConfig() {
    return 0;
}

sub checkArgs {
    if ($g_help || $g_cmdarg eq "") {
        usage();
    }
    # first figure out what actions were requested
    my @cmds = split(/,/, $g_cmdarg);
    for my $cmd (@cmds) {
        if ($cmd eq "readdb") {
            $g_doreaddb = 1;
        } elsif ($cmd eq "checkdb") {
            $g_docheckdb = 1;
        } elsif ($cmd eq "gensql") {
            $g_dogensql = 1;
        } elsif ($cmd eq "runsql") {
            $g_dorunsql = 1;
        } elsif ($cmd eq "fixdb") {
            $g_doreaddb = $g_dogensql = $g_dorunsql = 1;
        } elsif ($cmd eq "test") {
            $g_dotest = 1;
        } elsif ($cmd eq "verifycfg") {
            $g_doverifycfg = 1;
        } else {
            usage("Unrecognized command:  $cmd");
        }
    }
    if (($g_docheckdb && ($g_dogensql || $g_dorunsql)) ||
        ($g_doreaddb && $g_dorunsql && !$g_dogensql) ||
        ($g_dotest && ($g_dogensql || $g_doreaddb || $g_dorunsql || $g_docheckdb)) ||
        ($g_doverifycfg && ($g_dogensql || $g_doreaddb || $g_dorunsql || $g_docheckdb || $g_dotest))
        ) {
        usage("Inconsistent commands specified - $g_cmdarg");
    }
    
    if ($g_removeutc && ($g_dorunsql ||$g_dogensql)) {
        if (!areTheySure("Removeutc removes all UTC information from the history table,\nconverting all dates to database server local time.  Are you sure?")) {
            warning("Removeutc not confirmed, exiting");
            exit(1);
        }
    }
    if (! $g_fmt) {
        $g_fmt = $FMT_VALUES;
    } elsif ($g_fmt ne $FMT_SQL && $g_fmt ne $FMT_VALUES) {
        usage("-fmt must be $FMT_VALUES or $FMT_SQL but is $g_fmt");
    }

    if ($g_dotest) {
        # use require so that customers do not need the test module
        require TimeChangeTests;
        my $teststat = runtests($g_fmt);
        if (!$teststat) {
            info("No errors detected");
        }
        exit $teststat;
    }

    # removeutc does not use the config file, just the info read from the database.
    if ((! $g_removeutc && ($g_docheckdb || $g_dogensql)) ||
        $g_doverifycfg || $g_cfgfile) {
        checkInputFile($g_cfgfile, "You must specify a configuration file for command $g_cmdarg");
        # process config file
        if ($g_doverifycfg && $g_verb < 2) {
            $g_verb = 2;
        }
        my @errs = ();
        if (readConfigFile($g_cfgfile, \@errs)) {
            unshift @errs, "Error processing config file $g_cfgfile";
            err(join("\n", @errs));
        }
    }

    # Make sure that input files were specified, if required, and that
    # they exist.
    if (!$g_doreaddb) {
        # If we are not reading the db, output files from reading
        # the db must be specified
        if ($g_docheckdb || $g_dogensql) {
            checkInputFile($g_histfile, "The history file is required");
            checkInputFile($g_idblockfile, "The id block file is required");
        }
    }
    if ($g_dorunsql && ! $g_dogensql) {
        checkInputFile($g_sqlfile, "The repair command file is required");
    }
    # Make sure that output files were specified
    if ($g_doreaddb) {
        checkOutputFile($g_histfile, "The history file is required");
        checkOutputFile($g_idblockfile, "The id block file is required");
    }
    if ($g_dogensql) {
        checkOutputFile($g_sqlfile, "The repair command file is required");
    }

    # If we are connecting to the db, we need pdsql arguments
    if ($g_doreaddb || $g_dorunsql) {
        if (! $g_pdsqlargs) {
            usage("Command $g_cmdarg requires -pdsqlargs");
        }
    }
}

sub areTheySure {
    my $msg = shift;
    print "$msg  [no|yes]\n";
    my $confirm = <STDIN>;
    chomp $confirm;
    return lc($confirm) eq "yes";
}

# Figure out which db vendor we are using
sub whatdb {
    my @args = split(" ", $g_pdsqlargs);
    # < not <=
    for (my $i = 0; $i < $#args; $i++) {
        if ($args[$i] eq "-v") {
            $g_dbvendor = $args[$i+1];
            last;
        }
    }
    # The legal vendor strings are in bugbert/src/pdsql/main.cpp
    # Do sql server last since ss is a substring of access.
    if (($g_dbvendor =~ s/.*access.*/access/i) ||
        ($g_dbvendor =~ s/.*db2.*/db2/i) ||
        ($g_dbvendor =~ s/.*ora.*/oracle/i) ||
        ($g_dbvendor =~ s/.*(ss|sql_server).*/sql_server/i)) {
        info("DB vendor is $g_dbvendor");
    } else {
        err("Unable to determine database vendor from $g_pdsqlargs");
    }
}

# Make sure that datetimes read from the database are formatted as
# yyyy-mm-dd hh24:mi:ss
sub vendorTime {
    my $fld = shift;
    my $expr = "";
    if ($g_dbvendor eq "access") {
        #$expr = "Format($fld, 'yyyy-mm-dd hh:mm:ss')";
        # access doesn't support ifnull and isnull doesn't work the same way as
        # ifnull, so punt
        return $fld;
    } elsif ($g_dbvendor eq "db2" || $g_dbvendor eq "oracle") {
        $expr = "to_char($fld, 'yyyy-mm-dd hh24:mi:ss')";
    } else {
        # sql server
        $expr = "convert(char(19), $fld, 120)";
    }
    my $fldonly;
    ($fldonly = $fld) =~ s/.*\.//;
    $expr = "{fn ifnull($expr, 'NULL')} as $fldonly";
    return $expr;
}

# Simple test for whether the db info is plausible.
# If a user specifies a master database instead of a user db to read from,
# pdsql does not return non-zero status, so this is a crude way to catch
# that case.
sub checkDbInfo {
    my $dbinfo = shift;
    $dbinfo =~ s/^#*\s*//;
    if ($dbinfo =~ m/^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}/i) {
        # begins with a uuid, assume it's ok
        return 1;
    }
    return 0;
}

sub dbInfoQuery {
    return "select t2.working_master, t1.clan, t1.family, t2.schema_name, " .
        "t1.name as site_name, t1.dbid, t2.site_uuid from " .
        "ratl_replicas t1, dbglobal t2 where " .
        "t1.uuid = t2.site_uuid;";
}

sub readFromDB {
    my $dbv = whatdb();
    my $alloc_tm = vendorTime("t1.alloc_tm");
    my $action_ts = vendorTime("t1.action_timestamp");
    my $expired_ts = vendorTime("t1.expired_timestamp");
    my $dbinfoquery = dbInfoQuery();
    my @cmds = (
                "pdsql $g_pdsqlargs > " . quotefn($g_idblockfile),
                "pdsql $g_pdsqlargs > " . quotefn($g_histfile)
                );
    my @queries = (
                   "select t1.base, t1.limit, t2.name, t1.type, $alloc_tm from " .
                   "ratl_id_blocks t1, ratl_replicas t2 " .
                   "where t1.type = 3 and t1.replica = t2.dbid order by t2.name, t1.base;",
                   "select t1.dbid, t1.entity_dbid, $action_ts, $expired_ts, " .
                   "t1.entitydef_name, t1.comments from history t1 " .
                   "order by entitydef_id, entity_dbid, action_timestamp desc, dbid desc;"
                   );

    my @msgs = (
                "id block and site information",
                "history information"
                );
    my $tm = scalar(localtime);
    prtmsg("Starting reads at $tm");
    my $ks = 0;
    
    for (my $i = 0; $i <= $#cmds; $i++) {
        my @args = (
                    $dbinfoquery,
                    $queries[$i],
                    "quit;"
                    );
        $ks = runcmd($cmds[$i], \@args);
        if ($ks) {
            err("Could not read $msgs[$i] from database");
        } else {
            prtmsg("Read $msgs[$i] from database successfully");
        }
        $tm = scalar(localtime);
        prtmsg("Read finished at $tm");
    }
    return $ks;
}

# The 2 timestamp fields can either be NULL or contain an embedded blank.
# The comments field is whatever is left at the end of the line.  It can be 
# NULL or contain embedded blanks.  E.g. AP1 1, EP1 1
# dbid, entity_dbid, action_timestamp, expired_timestamp, entitydef_name, comments
sub splitHistoryRow {
    my $r = shift;
    my @row = split(" ", $r);
    # action_timestamp must be first because it is selected first in the query
    for my $idx ($ACTION_TIMESTAMP_IDX, $EXPIRED_TIMESTAMP_IDX) {
        if ($row[$idx] ne 'NULL') {
            my $secondhalf = splice(@row, $idx + 1, 1);
            $row[$idx] .= " $secondhalf";
        }
    }
    my @comm = splice(@row, $COMMENTS_IDX);
    $row[$COMMENTS_IDX] = join(" ", @comm);
    return @row;
}

# Each row from the history table is represented by an array.  The
# first n elements are the values read from the table.  The first
# interesting row in the file contains the column names. 
sub checkAndGen {
    # If $gen is true, generate repair output, otherwise just report
    # whether there are updates needed.
    my $gen = shift;
    my $hfile = shift;
    my $sfile = shift;
    my $rchanged = shift;  # used by tests only
    open HF, "<$hfile" or die "Cannot open $hfile for reading: $!\n";

    my $sqlhandle;
    if ($gen) {
        open($sqlhandle,">$sfile") or die "Cannot open $sfile for writing: $!\n";
    }
    my $saw_hdrs = 0;
    my $cur_dbid = 0;
    my $cur_entity_dbid = 0;
    my $cur_act_ts = "";
    my $cur_exp_ts = "";
    my $cur_entity_def_name = "";
    my $cur_comments = "";
    my $prev_entity_dbid = $UNINIT;
    my $prev_entity_def_name = "";
    my $LIM = 20;
    $g_history_dbinfo = "";
    resetStats();
    my @cur_group = ();
    while (<HF>) {
        next if (m/^>/);     # pdsql prompt lines
        next if (m/^\s*$/);  # blank lines
        chomp;
        s/^\s+//;  # strip leading whitespace
        s/\s+$//;  # strip trailing whitespace
        if ($saw_hdrs != 2 && $. > $LIM) {
            # Don't read the whole file before giving up
            err("Did not find database info in first $LIM lines of $hfile");
        } elsif ($saw_hdrs == 0) {
            $saw_hdrs = 1;
            next;
        } elsif ($saw_hdrs == 1) {
            if ($g_history_dbinfo eq "") {
                ($g_history_dbinfo = $_) =~ s/\s+/ /g;
                if ($g_history_dbinfo eq $g_idblock_dbinfo) {
                    if ($gen) {
                        print $sqlhandle "# $g_history_dbinfo\n";
                        set_trans_mode($sqlhandle);
                    }
                } else {
                    my $msg = <<EOS;
History file and id block file are not from the same database:
  $g_history_dbinfo
  $g_idblock_dbinfo
EOS
                    err($msg);
                }
            } else {
                $saw_hdrs = 2;
            }
            next;
        }

        # When grouping rows, put row number from the file ($.) on the
        # list for use in error reporting.
        ($cur_dbid, $cur_entity_dbid, $cur_act_ts, $cur_exp_ts, $cur_entity_def_name, $cur_comments) =  splitHistoryRow($_);
        # history has a row with dbid 0, other fields null
        next if ($cur_dbid == 0);
        # ratlc had over 1400 rows with entity_dbid = 0.  Skip these.
        # Could warn about them.
        next if ($cur_entity_dbid == 0);

        if ($cur_dbid eq $NULLSTR || $cur_entity_dbid eq $NULLSTR ||
            $cur_entity_def_name eq $NULLSTR) {
            err("Unexpected null value(s) at line $. in $hfile");
        }
        if ($cur_act_ts eq $NULLSTR) {
            $cur_act_ts = "";
        }
        if ($cur_exp_ts eq $NULLSTR) {
            $cur_exp_ts = "";
        }
        if ($cur_comments eq $NULLSTR) {
            $cur_comments = "";
        }
        # If this is not the first row with data, and either the entity_dbid
        # or the entity_def_name differs from the previous row, then
        # process the current group of rows, which is the complete set
        # for this entity.
        if (($prev_entity_dbid != $UNINIT) &&
            ($cur_entity_dbid != $prev_entity_dbid ||
             $cur_entity_def_name != $prev_entity_def_name)) {
            proc_one_record(\@cur_group, $sqlhandle);
            @cur_group = ();
        }
        push @cur_group, [$cur_dbid, $cur_entity_dbid, $cur_act_ts, $cur_exp_ts, $cur_entity_def_name, $cur_comments, $.];
        $prev_entity_dbid = $cur_entity_dbid;
        $prev_entity_def_name = $cur_entity_def_name;
    }

    close HF;
    if ($#cur_group >= 0) {
        proc_one_record(\@cur_group, $sqlhandle);
    }
    commit_trans($sqlhandle);
    if ($sqlhandle && $g_fmt eq $FMT_SQL && $g_rowschanged > 0) {
        print $sqlhandle "quit;\n";
        close($sqlhandle);
    }

    summarize_stats();
    if (defined $rchanged) {
        $$rchanged = $g_rowschanged;
    }
    return 0;
}

# This routine reads database info from the database being updated and
# the file containing the repair commands.  If the first n fields
# don't match, the database is not updated.  One could imagine more
# stringent checking, like making sure that the site name and site
# uuid appear in ratl_replicas in the local database.  Even if you are
# running fixdb (ie doing all 3 operations in one invocation of the
# script), this verification is probably a good idea.  A file might
# have been unwritable, or some other error might not have been
# properly detected, and this will act as a last line of defense when
# a file was supposed to be updated but wasn't.
#
# It also makes sure that the -fmt flag used when running the sql matches
# the value used when generating the sql.  Could just set $g_fmt correctly.
sub verifyDBMatches {
    my $filefmt = "undefined";
    open SQL, "<$g_sqlfile" or die "cannot open $g_sqlfile for reading";
    while (<SQL>) {
        next if m/autocommit/;
        if (s/^#\s*//) {
            chomp;
            $g_sql_dbinfo = $_;
        } elsif ($g_sql_dbinfo ne "") {
            if (m/update history/i) {
                $filefmt = $FMT_SQL;
            } else {
                $filefmt = $FMT_VALUES;
            }
            last;
        }
    } 
    close SQL;
    if ($filefmt ne $g_fmt) {
        err("-fmt argument is '$g_fmt', but the format of $g_sqlfile is '$filefmt'.\n  Check to make sure that the file is in the correct format.");
    }
    my $dbinfoquery = dbInfoQuery();
    my $fn = tmpnam();
    my $cmd = "pdsql $g_pdsqlargs > " . quotefn($fn);
    my @sqlcmds = (
                   "$dbinfoquery",
                   "quit;"
                   );
    my $ks = runcmd($cmd, \@sqlcmds);
    my $saw_hdrs = 0;
    open DBI, "<$fn" or die "Cannot open $fn for reading";
    while (<DBI>) {
        next if (m/^>/);     # pdsql prompt lines
        next if (m/^\s*$/);  # blank lines
        chomp;
        s/^\s+//;  # strip leading whitespace
        s/\s+$//;  # strip trailing whitespace
        s/\s+/ /g; # turn all white space into single blanks
        # skip column headers and store dbinfo
        if ($saw_hdrs == 0) {
            $saw_hdrs = 1;
        } else {
            ($g_fixup_dbinfo = $_) =~ s/\s+/ /g;
            last;
        }
    }
    close DBI;
    unlink($fn);
    if (!$g_fixup_dbinfo || !$g_sql_dbinfo) {
        err("Database info missing from sql file or cannot be read from database");
    } else {
        my @fixdbflds = split(" ", $g_fixup_dbinfo);
        my @sqlflds = split(" ", $g_sql_dbinfo);
        for (my $i = 0; $i < $MATCHES_NEEDED; $i++) {
            if ($sqlflds[$i] ne $fixdbflds[$i]) {
                my $msg = <<EOS;
Database info mismatch between repair sql file and database being repaired:
  $g_fixup_dbinfo
  $g_sql_dbinfo
EOS
                err($msg);
            }
        }
    }
}

# Actually fix the database
sub repairDB {
    verifyDBMatches();
    my $cmd = "pdsql $g_pdsqlargs";
    my @sqlcmds = (
                "update_history $g_sqlfile;",
                "quit;"
                );
    my $ks = 0;
    if ($g_fmt eq $FMT_VALUES) {
        runcmd($cmd, \@sqlcmds, undef);
    } else {
        runcmd($cmd, undef, $g_sqlfile);
    }
    prtmsg("Repair command completed");
    return $ks;
}

# This is the main processing routine for both checking the history
# table and generating repair commands.

# This routine is called with a ref to an array of refs to anonymous
# arrays.  The array represents the set of rows from the history table for a
# particular record (defect, project, etc.), sorted by the current
# values of action_timestamp descending. Each anonymous array
# corresponds to the values in one history table row.

# Multiple passes are made:
#   1) Adjust action_timestamps for rows with no action_timestamp adjustment
#      comment.
#   2) Resort by new action_timestamps if any were changed.
#   3) Fix expired_timestamps and update comments.
#   4) Generate repair commands if any are needed and if we are not in check
#      only mode.

# In addition to converting the time to utc, this will fix the
# expired_timestamp problems caused by a couple bugs in the core.  That is
# this supercedes the fixhistory script, which is dependent on sorting by
# action_timestamp.
# Note that in diagnostic messages, line number is the line in the
# file on disk, index is the array index after the re-sort.
sub proc_one_record {
    my $record_arr = shift;
    my $outputhandle = shift;

    my $this_entitydef_name = "";  # for debugging only
    my $this_entity_dbid = 0;      # for debugging only

    # pass 1 - adjust action_timestamps
    my $actts_changes = 0;
    for (my $i = 0; $i <= $#$record_arr; $i++) {
        my $row = $$record_arr[$i];
        if ($i == 0) {
            $this_entitydef_name = $$row[$ENTITYDEF_NAME_IDX];
            $this_entity_dbid = $$row[$ENTITY_DBID_IDX];
        }
        # Parse the existing comments field once and remember what was
        # found, represented as a bitmask, in a pseudo_column.
        $$row[$COMMENTS_FMT_IDX] = timestampCommentFmt($$row[$COMMENTS_IDX]);
        if ($$row[$COMMENTS_FMT_IDX] == $BOGUS_COMM) {
            my $msg = <<EOS;
Illegally formatted comment at line $$row[$LINENO_IDX], index $i.
  Dbid = $$row[$DBID_IDX], entity_dbid = $$row[$ENTITY_DBID_IDX],
  comments = '$$row[$COMMENTS_IDX]'.
EOS
            # Give up now.
            err($msg);
        }
        if ($g_removeutc) {
            if (($$row[$COMMENTS_FMT_IDX] & $ACT_COMM)) {
                my $new_act_ts = remove_utc($$row[$COMMENTS_IDX], $$row[$ACTION_TIMESTAMP_IDX]);
                $$row[$ACTION_TIMESTAMP_IDX] = $new_act_ts;
                $$row[$COMMENTS_IDX] = "";
                $$row[$ROW_CHANGES_IDX] |= ($ACT_TS_CHANGE | $COMM_CHANGE);
                # Note that some action_timestamp changed so we need to re-sort
                $actts_changes++;
            }
        } else {
            if (($$row[$COMMENTS_FMT_IDX] & $ACT_COMM)) {
                # Save existing action timestamp comment for possible use in
                # setting expired timestamp comment.
                $$row[$ACT_TS_COMM_IDX] = extractActionComment($$row[$COMMENTS_IDX]);
            } else {
                # This row does not contain the action_timestamp conventional
                # comment, so convert it now, save the adjustment actually applied.
                # To do the conversion you need to pass the dbid and the
                # current action timestamp.
                my $new_act_ts;
                my $tz_offset;
                ($new_act_ts, $tz_offset) = fixtime_for_dbid($$row[$DBID_IDX], $$row[$ACTION_TIMESTAMP_IDX]);

                $$row[$ACTION_TIMESTAMP_IDX] = $new_act_ts;
                # Record the action_timestamp conventional comment
                $$row[$ACT_TS_COMM_IDX] = "AC1 $tz_offset";
                # Remember type of change made for this row
                $$row[$ROW_CHANGES_IDX] |= ($ACT_TS_CHANGE | $COMM_CHANGE);
                # Note that some action_timestamp changed so we need to re-sort
                $actts_changes++;
            }
        }
    }

    # Pass 2 - resort by action timestamps
    if ($actts_changes) {
        # Now re-sort by action_timestamp because at least 1 of them
        # changed.  This can produce lots of output for a big
        # database, so be careful about turning it on.
        if ($g_verb >= $DEF_INFO_VERB) {
            info("Re-sorting $this_entitydef_name $this_entity_dbid");
        }
        @$record_arr = sort ts_sort @$record_arr;
        # See if the sort changed anything by seeing if the line numbers
        # are still in decreasing order.  Avoid this expensive check by
        # checking $g_verb before doing it.
        if ($g_verb >= $DEF_INFO_VERB) {
            for (my $i = 0; $i < $#$record_arr; $i++) {
                my $r1 = $$record_arr[$i];
                my $r2 = $$record_arr[$i + 1];
                if ($$r1[$LINENO_IDX] != $$r2[$LINENO_IDX] - 1) {
                    info("Re-sort changed order for $this_entitydef_name $this_entity_dbid");
                    last;
                }
            }
        }
    }

    # Pass 3 - fix comments and expired timestamps that need fixing
    # For any row with a change, $$row[$ROW_CHANGES_IDX] must be set!
    my $prev_action_ts = "";
    my $prev_action_ts_comm = "";
    for (my $i = 0; $i <= $#$record_arr; $i++) {
        my $row = $$record_arr[$i];
        if ($g_removeutc) {
            # zap the comments
            if ($$row[$COMMENTS_FMT_IDX] != $NO_COMM) {
                $$row[$COMMENTS_IDX] = "";
                $$row[$ROW_CHANGES_IDX] |= $COMM_CHANGE;
            }
            if ($i == 0) {
                # non-null expired_timestamp?
                if ($$row[$EXPIRED_TIMESTAMP_IDX] ne "") {
                    $$row[$ROW_CHANGES_IDX] |= $EXP_TS_CHANGE;
                    $$row[$EXPIRED_TIMESTAMP_IDX] = "";
                }
            } else {
                # rows 2 - n
                if (! check_ts_consistency($prev_action_ts, $$row[$EXPIRED_TIMESTAMP_IDX])) {
                    $$row[$EXPIRED_TIMESTAMP_IDX] = $prev_action_ts;
                    $$row[$ROW_CHANGES_IDX] |= $EXP_TS_CHANGE;
                }
            }
        } else {
            if ($i == 0) {
                # first record of group
                if ($$row[$COMMENTS_FMT_IDX] & $EXP_COMM) {
                    # First record should not have an expired timestamp
                    # comment, because the expired_timestamp should be null.
                    my $msg = <<EOS;
Unexpected non-null expired_timestamp comment at line $$row[$LINENO_IDX], index $i.
  Dbid = $$row[$DBID_IDX], entity_dbid = $$row[$ENTITY_DBID_IDX],
  expired_timestamp = '$$row[$EXPIRED_TIMESTAMP_IDX]', comments = '$$row[$COMMENTS_IDX]'.
EOS
                        warning($msg);
                    # fix it now
                    $$row[$COMMENTS_IDX] = $$row[$ACT_TS_COMM_IDX];
                    $$row[$ROW_CHANGES_IDX] |= $COMM_CHANGE;
                } 
                if ($$row[$EXPIRED_TIMESTAMP_IDX] ne "") {
                    my $msg =  <<EOS;
Unexpected non-null expired_timestamp at line $$row[$LINENO_IDX], index $i.
  Expired_timestamp is non-null for the newest record in a set.
  Dbid = $$row[$DBID_IDX], entity_dbid = $$row[$ENTITY_DBID_IDX],
  Expired_timestamp = '$$row[$EXPIRED_TIMESTAMP_IDX]'.
EOS
                        info($msg, 2);
                    $$row[$EXPIRED_TIMESTAMP_IDX] = "";
                    $$row[$ROW_CHANGES_IDX] |= $EXP_TS_CHANGE;
                }
                if ($$row[$ROW_CHANGES_IDX] & $ACT_TS_CHANGE) {
                    # The action_timestamp itself was adjusted prior to
                    # the re-sort of the array.  Record the comments now
                    # that we are sure it's the first row.
                    $$row[$COMMENTS_IDX] = "$$row[$ACT_TS_COMM_IDX]";
                    # Mark that this row changed
                    $$row[$ROW_CHANGES_IDX] |= $COMM_CHANGE;
                }
            } else {
                # 2nd - nth rows
                if (!($$row[$COMMENTS_FMT_IDX] & $EXP_COMM)) {
                    $$row[$ROW_CHANGES_IDX] |= $COMM_CHANGE;
                } elsif ($$row[$EXPIRED_TIMESTAMP_IDX] eq "") {
                    # expired_timestamp not converted or null (shouldn't be null,
                    # but RATLC00762411 can cause this).
                    $$row[$ROW_CHANGES_IDX] |= $EXP_TS_CHANGE;
                } elsif (! check_ts_consistency($prev_action_ts, $$row[$EXPIRED_TIMESTAMP_IDX])) {
                    # Even though the old expired_timestamp was in utc, it didn't
                    # match the action_timestamp it was supposed to.  Warn about it
                    # and fix it up.  I don't expect this to happen but wouldn't
                    # be totally surprised if it did.
                    my $msg = <<EOS;
Fixing expired timestamp that was already in utc at line $$row[$LINENO_IDX], index $i.
  Dbid = $$row[$DBID_IDX], entity_dbid = $$row[$ENTITY_DBID_IDX],
  $prev_action_ts != $$row[$EXPIRED_TIMESTAMP_IDX]
EOS
                        warning($msg);
                    $$row[$ROW_CHANGES_IDX] |= $EXP_TS_CHANGE;
                } 
                # If there are any changes for this row, including a change
                # to the action_timestamp, update all the fields now.
                if ($$row[$ROW_CHANGES_IDX] != 0) {
                    $$row[$EXPIRED_TIMESTAMP_IDX] = $prev_action_ts;
                    my $exp_comm = actionCommToExpComm($prev_action_ts_comm);
                    $$row[$COMMENTS_IDX] = "$$row[$ACT_TS_COMM_IDX], $exp_comm";
                }
            }

            # check if we generated an ok comment and croak if we didn't
            my $commentFmt = timestampCommentFmt($$row[$COMMENTS_IDX]);
            if ($commentFmt == $NO_COMM || $commentFmt == $BOGUS_COMM ||
                (($commentFmt & $EXP_COMM) && $i == 0) || 
                (!($commentFmt & $EXP_COMM) && $i > 0)) {
                my $msg = <<EOS;
Comment is not in correct format at line $$row[$LINENO_IDX], index $i.
  Comment = $$row[$COMMENTS_IDX].
EOS
                    err($msg);
            }

            # Remember previous action timestamp conventional comment
            # for possible use when setting expired_timestamp of next
            # row.  Not needed for removeutc.
            $prev_action_ts_comm = $$row[$ACT_TS_COMM_IDX];
        }

        # Remember previous action timestamp for possible use when
        # setting expired_timestamp of next row.
        $prev_action_ts = $$row[$ACTION_TIMESTAMP_IDX];
    }

    # Pass 4 - write sql
    my $changed_rows = 0;
    for (my $i = 0; $i <= $#$record_arr; $i++) {
        my $row = $$record_arr[$i];
        if ($$row[$ROW_CHANGES_IDX] == 0) {
            $g_nochanges_rows{$this_entitydef_name}++;
        } else {
            $changed_rows++;
            $g_updated_rows{$this_entitydef_name}++;
            if ($outputhandle) {
                $g_update_count++;
                if ($g_fmt eq $FMT_SQL) {
                    my $sql = "update history set action_timestamp = '$$row[$ACTION_TIMESTAMP_IDX]', " .
                        "expired_timestamp = '$$row[$EXPIRED_TIMESTAMP_IDX]', " .
                        "comments = '$$row[$COMMENTS_IDX]' " .
                        "where dbid = $$row[$DBID_IDX]; " .
                        "# $$row[$ENTITYDEF_NAME_IDX] $$row[$ENTITY_DBID_IDX]";
                    $sql =~ s/''/NULL/g;
                    print $outputhandle "$sql\n";
                } else {
                    # Characters after the # are comments.
                    # The order in which the non-comments are printed is the
                    # same as the order of params in the update statement in
                    # pdscript.cpp.
                    # "update history set action_timestamp = ?,
                    # expired_timestamp = ?, comments = ? where dbid = ?"
                    printf($outputhandle "%s\t%s\t%s\t%s\t# %s %s\n", 
                           $$row[$ACTION_TIMESTAMP_IDX], $$row[$EXPIRED_TIMESTAMP_IDX], 
                           $$row[$COMMENTS_IDX], $$row[$DBID_IDX], 
                           $$row[$ENTITYDEF_NAME_IDX], $$row[$ENTITY_DBID_IDX]);
                }
            }
        }
    }
    if ($changed_rows) {
        $g_updated_recs{$this_entitydef_name}++;
    } else {
        $g_nochanges_recs{$this_entitydef_name}++;
    }
    # Include all updates for a particular record in one transaction
    if ($g_update_count > $COMMIT_FREQ) {
        commit_trans($outputhandle);
    }
}

sub reportResults {
    # info("Who knows what happened?");
}

# This routine needed to make sure that each test gets a clean start, not
# needed otherwise.
sub resetStats {
    $g_rowschanged = 0;
    %g_updated_recs = ();
    %g_updated_rows = ();
    %g_nochanges_recs = ();
    %g_nochanges_rows = ();
    %g_sitefixups = ();
}

# generate status messages for the check and gen phase
sub summarize_stats {
    my $recschanged = 0;
    my $recsunchanged = 0;
    my $rowsunchanged = 0;
    for my $k (keys %g_updated_recs) {
        $recschanged += $g_updated_recs{$k};
    }
    for my $k (keys %g_updated_rows) {
        # rowschanged global
        $g_rowschanged += $g_updated_rows{$k};
    }
    for my $k (keys %g_nochanges_recs) {
        $recsunchanged += $g_nochanges_recs{$k};
    }
    for my $k (keys %g_nochanges_rows) {
        $rowsunchanged += $g_nochanges_rows{$k};
    }

    info("Summary of changes needed:", 1);
    info("  Records needing update:      $recschanged", 1);
    info("  Records not needing update:  $recsunchanged", 1);
    info("  Rows needing update:         $g_rowschanged", 1);
    info("  Rows not needing update:     $rowsunchanged", 1);

    # Note that the dbid -> site lookup only happens if the action_timestamp
    # needs converting.  If there are other fixups done on a row, such as
    # updating the expired_timestamp, that will not be captured in this hash.
    info("Action timestamps needing update, by site where dbid was allocated:", 1);
    for my $k (sort keys %g_sitefixups) {
        info("  $k:  $g_sitefixups{$k}", 1);
    }
}

# $cmd is the command line
# $cmdin is an array ref of commands
# $cmdfile is a filename that already contains commands to run
sub runcmd {
    my $cmd = shift;
    my $cmdin = shift;
    my $cmdfile = shift;
    my $ks = 0;
    if (defined $cmdin) {
        # Write the commands to a temp file and use that as redirected
        # input.  Otherwise there appears to be no way to get exit status
        # from the program.  Originally I was opening a pipe to the
        # command and writing the commands to it but that didn't seem to
        # report when the program exited.
        my $tempfn = tmpnam();
        open TMP, ">$tempfn" or die "Cannot open $tempfn: $!";
        for my $sql (@$cmdin) {
            print TMP "$sql\n";
        }
        close TMP;

        $ks = system("$cmd <" . quotefn($tempfn));
        unlink $tempfn;
    } elsif (defined $cmdfile) {
        $ks = system("$cmd <" . quotefn($cmdfile));
    } else {
        err("missing argument to runcmd");
    }
    return $ks;
}

# When system is invoking the shell to handle meta-characters for
# redirection, filenames with blanks need quotes.  When you are
# opening files using open(">$fn"), for example, you cannot have
# quotes.
sub quotefn {
    my $fn = shift;
    if ($fn =~ m/\s/) {
        $fn = '"' . $fn . '"';
    }
    return $fn;
}

# Figure out the format of a comments field and return a bit mask
# describing it.  Could do more rigorous checking for legality.
sub timestampCommentFmt {
    my $comm = shift;
    my $res = $NO_COMM;
    if ($comm !~ m/^\s*$/) {
        if ($comm =~ m/^A/) {
            if ($comm =~ m/^A[CP]\d/) {
                $res |= $ACT_COMM;
            } else {
                $res = $BOGUS_COMM;
            }
        }
        if ($res != $BOGUS_COMM && $comm =~ m/,/) {
            if ($comm =~ m/,.*,/) {
                $res = $BOGUS_COMM;
            } elsif ($comm =~ m/, E[CP]\d/) {
                $res |= $EXP_COMM;
            } else {
                $res = $BOGUS_COMM;
            }
        }
        if ($res == 0) {
            $res = $BOGUS_COMM;
        }
    }
    return $res;
}

# Given a conventional comment, get just the action timestamp part.
# E.g. Given AC1 -05:00, EC1 +06:30, return AC1 -05:00
sub extractActionComment {
    my $comm = shift;
    my $res;
    ($res = $comm) =~ s/,.*//;
    return $res;
}

# Given a conventional comment, get just the action timestamp offset part.
# E.g. Given AC1 -05:00, EC1 +06:30, return -05:00
sub extractActionOffset {
    my $comm = shift;
    my $actcomm = extractActionComment($comm);
    my $res;
    ($res = $actcomm) =~ s/^.*\s//;
    return $res;
}

# Convert an action timestamp comment convert it to an expired
# timestamp comment.  E.g. AC1 -05:00 becomes EC1 -05:00
sub actionCommToExpComm {
    my $comm = shift;
    my $exp_comm;
    ($exp_comm = $comm) =~ s/^A/E/;
    return $exp_comm;
}

# Binary search.  We don't ship any perl modules that provide binary
# search so we have to provide our own.  Expects a ref to an array of
# array refs.  The first 2 elements of these arrays are the end points
# of the range.  Note it does inclusive comparisons on both endpoints.
# Returns count to caller for use in error messages and tests.
# Should take a callback function for doing the comparison, but since
# its use is limited to 2 places, avoid the extra overhead.
sub brangesrch {
    my $ar = shift;
    my $val = shift;
    my $sz = $#$ar;
    if ($sz < 0) {
        err("Empty array passed to brangesrch");
    }
    my $res = -1;
    my $cnt = 0;
    my $left = 0;
    my $right = $sz;
    # Limit the number of times through the loop to prevent infinite loops
    # in case of a bug.
    while ($cnt++ <= $sz) {
        if ($right < $left) {
            last;
        }
        my $idx = int(($left + $right) / 2);
        if ($idx < 0 || $idx > $sz) {
            err("Bogus value for index in brangesrch:  $idx");
        }
        if ($g_verb >= 3) {
            info("left = $left, idx = $idx, right = $right");
        }
        my $rec = $$ar[$idx];
        if ($val >= $$rec[0] && $val <= $$rec[1]) {
            $res = $idx;
            last;
        }
        if ($val > $$rec[1]) {
            $left = $idx + 1;
        } else {
            $right = $idx - 1;
        }
    }
    return ($res, $cnt);
}
